/*************************************************************
 * (C) 2001, Roberto A. Foglietta <robang@libero.it>
 *           Tutti i diritti riservati oppure GPL.
 *  
 * (C) 2001, Roberto A. Foglietta <robang@libero.it>
 *           Delivered under GPL as beta software,
 *           use at your own risk.
 *           
 * http://digilander.iol.it/robang
 *                    
 * rafobj.c : nuovo modulo gestione oggetti  
 * creazione: 2001-04-25 
 * modifiche: 2001-05-28 v0.1 
 * 						2001-05-29 v0.1.1 BUG FIX & TESTED
 * 						2001-05-30 v0.1.2 some GET call SLT
 * 						2001-06-22 v0.1.3 NEXT/PREV return NULL at the end of buf
 * 						2001-07-01 v0.1.4 SLT | SET add as local method
 *						2001-07-03 v0.1.5 SORT with qsort function
 *						2001-07-05 v0.1.6 SLT | RPL metod as replacement of selected member
 *						2001-07-05 v0.1.7	CPY metod works with NULL pAram too
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#ifdef ANJUTA_MAKEFILE
# include <gtk/gtk.h>
# include "pfunc.h"
#endif
#include "rafobj.h"

#define RAF_SET     0x01 

struct _raf_obj {	RAF_TYPE_OBJECT *obj;
                  long cod;
                  char del;
                };
typedef struct _raf_obj rafobj;

static int  (*usr_cmp)(const void *el1, const void *el2);
static int raf_obj_cmp(const void *el1, const void *el2);
static int raf_obj_cmp(const void *el1, const void *el2)
{
	/* Faster if remmed than disabled */
	//RAF_DEBUG_INIT{"raf_obj_cmp"};
	rafobj *e1, *e2;
	
	e1 = (rafobj *)el1;
	e2 = (rafobj *)el2;
	
	//RAF_DEBUG("cod:%ld,%ld\n", e1->cod, e2->cod);

	if(e1 == e2) return 0;       
  else if(e1 == NULL) return 1; 
	else if(e2 == NULL) return -1;
	else return usr_cmp( e1->obj, e2->obj );
}
										
const void *raf_obj_manage(int flag, void *pAram)
{
	RAF_DEBUG_INIT{"", "raf_obj_manage"};
	/* RAF 2001-05-23: questa funzione si comporta come 
	 * un C++ Object dove i flags fungono da metodi.
	 */
	static int     prima  = 1;
	static long    Freeze = 0;
	static long    NumObj = 0;
	static rafobj  List[RAF_MAX_NUM_OBJECT];
	long i, u = 0;
	void *v;

	/* RAF 2001-05-25: inizializzazione */
	if(prima) {
		for(i = 0; i < RAF_MAX_NUM_OBJECT; i++) 
			List[i].obj = NULL;
			List[i].cod = 0;
			List[i].del = 0;
	 	prima = 0;
	}
	
	switch(flag & 0xF0) {
		case RAF_OBJ_RPL:
		/* RAF 2001-07-05: replace selected member with a copy of passed obj */
			u = raf_object_get_ind;
			if(List[u].obj) 
				free(List[u].obj);
			List[u].obj = (RAF_TYPE_OBJECT *)raf_obj_manage(RAF_OBJ_CPY, pAram);
			//List[NumObj].cod = ;
			List[NumObj].del = 0;
			return (void *)u;
			break;
		

		case RAF_OBJ_ORD:
		/* RAF 2001-07-01: l'uso di qsort diventa necessario per
	  	 ordinare la lista; infatti quando il numero di elenti 
		   diventa molto grande il tempo di ordinamento cresce
		   solo come n*log(n) invece di circa n 
	 	 */
		usr_cmp = pAram;
		qsort(List, NumObj, (size_t)sizeof(rafobj), raf_obj_cmp);	
		return (void *)NumObj;
		break;

		case RAF_OBJ_ADD:
			/* RAF 2001-05-25: 
			 * genera un codice identificativo univoco 
			 */
			u = raf_object_get_unq;
			
			/* RAF 2001-05-28: 
			 * genera un copia dell'oggetto passatogli
			 * TODO: il cast non viene eseguito ma solo 
			 * inizializzato anche se per ora e` solo 
			 * una costante 
			 */
			List[NumObj].obj = (RAF_TYPE_OBJECT *)raf_obj_manage(RAF_OBJ_CPY, pAram);
			List[NumObj].cod = u;
			List[NumObj].del = 0;
			NumObj++;

			/* RAF 2001-05-25: 
			 * ritorna il codice identificativo univoco 
			 */
			RAF_DEBUG("RAF_OBJ_ADD aggiunto %ld in %ld\n", u, NumObj);
			return (void *)u;
			break;
			
		case RAF_OBJ_CPY:
			/* RAF 2001-07-05: 
			 * Se passo un parametro nullo prendo il selezionato
			 */
			if(pAram == NULL)	
				return NULL;
					
			/* RAF 2001-05-28:
		 	 * genera una nuova istanza dell'oggetto 
		 	 */
			v = (void *)malloc(sizeof(RAF_TYPE_OBJECT));
			v = memcpy(v, pAram, sizeof(RAF_TYPE_OBJECT));		
			return (void *)v;
			break;
			
		case RAF_OBJ_RMV:
			/* RAF 2001-05-28:
			 * In alcune situazioni il contenuto deve essere 
			 * accessibile per un certo periodo anche se nel 
			 * contempo alcuni dovrebbero essere eliminati.
			 * Con il freeze/thaw i delete sono posticipati
			 */
			switch(flag & 0x0F) {
				case RAF_FREEZE:
					Freeze++;
					if(!Freeze) 
					 Freeze = (unsigned)-1;
					return (void *)Freeze;
					break;

				case RAF_ALL:
					/* RAF 2001-07-01 */
					if(Freeze) 
						for(i = 0; i < NumObj; i++)
							List[i].del = 1;
					else
						for(i = 0; i < NumObj; i++) 
							raf_obj_manage(RAF_OBJ_RMV | RAF_IND, (void *)i);
					return (void *)NULL;
					break;

				case RAF_THAW:
					if(Freeze > 0)
					 Freeze--;
					/* RAF 2001-05-28:
					 * Elimino quelli che sono stati  
					 * accodati durante il freezing time
					 */
					if(!Freeze) {
						for(i = 0; i < NumObj; i++) 
							if(List[i].del)
								raf_obj_manage(RAF_OBJ_RMV | RAF_IND, (void *)i);
					}
					return (void *)Freeze;
					break;
			}
			
			switch(flag & 0x0F) {
				case RAF_COD:
					/* RAF 2001-05-28: 
				 	 * Lo rimuovo per codice identificativo
			 		 */
					u = (long)pAram;
					if( raf_obj_manage(RAF_OBJ_SLT | RAF_COD, pAram) ) 
						raf_obj_manage(RAF_OBJ_RMV, NULL);
					else { 
						RAF_DEBUG("RAF_OBJ_RMV non trovato %ld\n", u); 
					}
					break;

				case RAF_IND:
					/* RAF 2001-05-28: 
				 	 * Lo rimuovo per indice (dall'esterno)
			 		 */
					i = (unsigned)pAram;
					if( raf_obj_manage(RAF_OBJ_SLT | RAF_COD, (void *)List[i].cod) ) 
						raf_obj_manage(RAF_OBJ_RMV, NULL);
					else { 
						RAF_DEBUG("RAF_OBJ_RMV non trovato %ld\n", u); 
					}
					break;
				
				default:
					i = raf_object_get_ind;
					if(!Freeze)	{
						/* RAF 2001-05-28:
					 	 * Elimino _veramente_ quello selezionato
					 	 */
						RAF_DEBUG("RAF_OBJ_RMV rimozione %ld cod. %ld\n", i, List[i].cod);
						free(List[i].obj);
						for(; i < NumObj-1; i++) {
							List[i].obj = List[i+1].obj;
							List[i].cod = List[i+1].cod;
							List[i].del = List[i+1].del;
						}
						List[i].obj = NULL;
						List[i].cod = 0;
						List[i].del = 0;
						NumObj--;
					}	else {
						/* RAF 2001-05-28: 
				 		 * Invece di eliminare, accodo l'elemento
			 	 		 */
						RAF_DEBUG("RAF_OBJ_RMV accodato per l'eliminazione\n");
						List[i].del = 1;
					}
					break;
			} 
			return (void *)NumObj;	
			break;
	}

	switch(flag & 0xF0) {
		/* RAF 2001-05-25: per maggiore astrazione l'indice
		 * e` disponibile solo per i metodi SP_SLT e SP_GET 
		 */
 		static long Select = 0;
		
		case RAF_OBJ_SLT:
			RAF_DEBUG("RAF_OBJ_SLT from %ld\n", Select);
			switch(flag & 0x0F) {
				case RAF_FIRST:
					Select = 0;
					break;
				
				case RAF_PREV:
					if(Select > 0) Select--;
					/* RAF 2001-06-22 serve per il while!NULL PREV*/
					else return NULL;		
					break;

				case RAF_NEXT:
					if(Select+1 < NumObj) Select++;
					/* RAF 2001-06-22 serve per il while!NULL NEXT*/
					else return NULL;		
					break;
					
				case RAF_LAST:
					Select = NumObj-1;
					break;
				
				case RAF_COD:
					/* RAF 2001-05-28: (long)pAram
					 * Seleziona un elemento ricercandolo per codice 
					 */ 
					u = (long)pAram;
					for(i = 0; i < NumObj; i++)
						if(List[i].cod == u) 
							break;
					if(i < NumObj) {
							RAF_DEBUG("RAF_OBJ_SLT trovato %lu in pos. %ld/%ld (0x%lx)\n", 
												u, i, NumObj, (long)List[i].obj);
							Select = i;
					} else { 
							RAF_DEBUG("RAF_OBJ_SLT non trovato %ld\n", u); 
							return NULL;
					}
					break;
			}
			v = (void *)(List[Select].obj);
			break;
		
		case RAF_OBJ_GET:
			switch(flag & 0x0F) {
				case RAF_IND:
					RAF_DEBUG("RAF_OBJ_GET | RAF_IND\n");
					v = (void *)Select;
					break;
					
				case RAF_NUM:
					RAF_DEBUG("RAF_OBJ_GET | RAF_NUM\n");
					v = (void *)NumObj;
					break;
					
				case RAF_COD:
					RAF_DEBUG("RAF_OBJ_GET | RAF_COD\n");
					v = (void *)(List[Select].cod);
					break;

				case RAF_FREEZE:
				case RAF_THAW:
					RAF_DEBUG("RAF_OBJ_GET | RAF_FREEZE/THAW\n");
					v = (void *)Freeze;
					break;

				case RAF_UNQ: {
					long max = 0;
					RAF_DEBUG("RAF_OBJ_GET | RAF_UNQ\n");
					for(i = 0; i < NumObj; i++)
						if(max <= List[i].cod) 
							max = List[i].cod;
					max++;
					RAF_DEBUG("Numero unico: %04lu\n", max);
				  v = (void *)max;
					break;
				}
				break;
					
				case RAF_FIRST:
				case RAF_PREV :
				case RAF_NEXT :
				case RAF_LAST :
					RAF_DEBUG("RAF_OBJ_GET -> RAF_OBJ_SLT\n");
					/* RAF 2001-05-30: viene naturale usare GET invece di SLT 
					 * e poiche` non vale la pena di fare i  difficili redirigo 
					 * la chiamata GET alla SLT che ha i sub_flag chiamati.
					 */ 
					v = (void *)raf_obj_manage(RAF_OBJ_SLT | (flag & 0x0F), pAram);
					break;
				
				default:
					RAF_DEBUG("RAF_OBJ_GET default case\n");
					v = List[Select].obj;
					break;
			}
			break;
					
		default:
			v = List[Select].obj;
			break;
	}		
	
	return v;
}


